@@ -15,12 +15,14 @@
 module Cryptol.ModuleSystem.Base where
 
 import Cryptol.ModuleSystem.Env (DynamicEnv(..), deIfaceDecls)
+import Cryptol.ModuleSystem.Fingerprint
 import Cryptol.ModuleSystem.Interface
 import Cryptol.ModuleSystem.Monad
 import Cryptol.ModuleSystem.Name (Name,liftSupply,PrimMap)
 import Cryptol.ModuleSystem.Env (lookupModule
                                 , LoadedModule(..)
-                                , meCoreLint, CoreLint(..))
+                                , meCoreLint, CoreLint(..)
+                                , ModulePath(..), modulePathLabel)
 import qualified Cryptol.Eval                 as E
 import qualified Cryptol.Eval.Value           as E
 import           Cryptol.Prims.Eval ()
@@ -44,17 +46,16 @@ import Cryptol.Utils.PP (pretty)
 import Cryptol.Utils.Panic (panic)
 import Cryptol.Utils.Logger(logPutStrLn, logPrint)
 
-import Cryptol.Prelude (writePreludeContents)
+import Cryptol.Prelude (preludeContents)
 
 import Cryptol.Transform.MonoValues (rewModule)
 
-import Control.DeepSeq
 import qualified Control.Exception as X
 import Control.Monad (unless,when)
+import qualified Data.ByteString as B
 import Data.Maybe (fromMaybe)
 import Data.Monoid ((<>))
-import           Data.Text(Text)
-import qualified Data.Text.IO as T
+import Data.Text.Encoding (decodeUtf8')
 import System.Directory (doesFileExist, canonicalizePath)
 import System.FilePath ( addExtension
                        , isAbsolute
@@ -67,7 +68,7 @@ import qualified System.IO.Error as IOE
 import qualified Data.Map as Map
 
 import Prelude ()
-import Prelude.Compat
+import Prelude.Compat hiding ( (<>) )
 
 
 -- Renaming --------------------------------------------------------------------
@@ -105,23 +106,39 @@ noPat a = do
 
 -- Parsing ---------------------------------------------------------------------
 
-parseModule :: FilePath -> ModuleM (P.Module PName)
+parseModule :: ModulePath -> ModuleM (Fingerprint, P.Module PName)
 parseModule path = do
 
-  e <- io $ X.try $ do
-    bytes <- T.readFile path
-    return $!! bytes
-  bytes <- case (e :: Either X.IOException Text) of
+  bytesRes <- case path of
+                InFile p -> io (X.try (B.readFile p))
+                InMem _ bs -> pure (Right bs)
+
+  bytes <- case bytesRes of
     Right bytes -> return bytes
-    Left exn | IOE.isDoesNotExistError exn -> cantFindFile path
-             | otherwise                   -> otherIOError path exn
+    Left exn ->
+      case path of
+        InFile p
+          | IOE.isDoesNotExistError exn -> cantFindFile p
+          | otherwise                   -> otherIOError p exn
+        InMem p _ -> panic "parseModule"
+                       [ "IOError for in-memory contetns???"
+                       , "Label: " ++ show p
+                       , "Exception: " ++ show exn ]
+
+  txt <- case decodeUtf8' bytes of
+    Right txt -> return txt
+    Left e    -> badUtf8 path e
 
   let cfg = P.defaultConfig
-              { P.cfgSource  = path
-              , P.cfgPreProc = P.guessPreProc path
+              { P.cfgSource  = case path of
+                                 InFile p -> p
+                                 InMem l _ -> l
+              , P.cfgPreProc = P.guessPreProc (modulePathLabel path)
               }
-  case P.parseModule cfg bytes of
-    Right pm -> return pm
+
+  case P.parseModule cfg txt of
+    Right pm -> let fp = fingerprint bytes
+                in fp `seq` return (fp, pm)
     Left err -> moduleParseError path err
 
 
@@ -132,25 +149,26 @@ loadModuleByPath :: FilePath -> ModuleM T.Module
 loadModuleByPath path = withPrependedSearchPath [ takeDirectory path ] $ do
   let fileName = takeFileName path
   foundPath <- findFile fileName
-  pm <- parseModule foundPath
+  (fp, pm) <- parseModule (InFile foundPath)
   let n = thing (P.mName pm)
 
   -- Check whether this module name has already been loaded from a different file
   env <- getModuleEnv
   -- path' is the resolved, absolute path, used only for checking
   -- whether it's already been loaded
-  path' <- io $ canonicalizePath foundPath
+  path' <- io (canonicalizePath foundPath)
+
   case lookupModule n env of
     -- loadModule will calculate the canonical path again
-    Nothing -> doLoadModule (FromModule n) foundPath pm
+    Nothing -> doLoadModule (FromModule n) (InFile foundPath) fp pm
     Just lm
      | path' == loaded -> return (lmModule lm)
      | otherwise       -> duplicateModuleName n path' loaded
-     where loaded = lmCanonicalPath lm
+     where loaded = lmModuleId lm
 
 
 -- | Load a module, unless it was previously loaded.
-loadModuleFrom :: ImportSource -> ModuleM (FilePath,T.Module)
+loadModuleFrom :: ImportSource -> ModuleM (ModulePath,T.Module)
 loadModuleFrom isrc =
   do let n = importedModule isrc
      mb <- getLoadedMaybe n
@@ -159,16 +177,18 @@ loadModuleFrom isrc =
        Nothing ->
          do path <- findModule n
             errorInFile path $
-              do pm <- parseModule path
-                 m  <- doLoadModule isrc path pm
+              do (fp, pm) <- parseModule path
+                 m        <- doLoadModule isrc path fp pm
                  return (path,m)
 
 -- | Load dependencies, typecheck, and add to the eval environment.
-doLoadModule :: ImportSource ->
-              FilePath ->
-              P.Module PName ->
-              ModuleM T.Module
-doLoadModule isrc path pm0 =
+doLoadModule ::
+  ImportSource ->
+  ModulePath ->
+  Fingerprint ->
+  P.Module PName ->
+  ModuleM T.Module
+doLoadModule isrc path fp pm0 =
   loading isrc $
   do let pm = addPrelude pm0
      loadDeps pm
@@ -179,16 +199,17 @@ doLoadModule isrc path pm0 =
 
      -- extend the eval env, unless a functor.
      unless (T.isParametrizedModule tcm) $ modifyEvalEnv (E.moduleEnv tcm)
-     canonicalPath <- io (canonicalizePath path)
-     loadedModule path canonicalPath tcm
+     loadedModule path fp tcm
 
      return tcm
   where
   optionalInstantiate tcm
-    | isParamInstModName (importedModule isrc) && T.isParametrizedModule tcm =
-      case addModParams tcm of
-        Right tcm1 -> return tcm1
-        Left xs    -> failedToParameterizeModDefs (T.mName tcm) xs
+    | isParamInstModName (importedModule isrc) =
+      if T.isParametrizedModule tcm then
+        case addModParams tcm of
+          Right tcm1 -> return tcm1
+          Left xs    -> failedToParameterizeModDefs (T.mName tcm) xs
+      else notAParameterizedModule (T.mName tcm)
     | otherwise = return tcm
 
 
@@ -215,8 +236,9 @@ importIfaces is = mconcat `fmap` mapM importIface is
 moduleFile :: ModName -> String -> FilePath
 moduleFile n = addExtension (joinPath (modNameChunks n))
 
+
 -- | Discover a module.
-findModule :: ModName -> ModuleM FilePath
+findModule :: ModName -> ModuleM ModulePath
 findModule n = do
   paths <- getSearchPath
   loop (possibleFiles paths)
@@ -225,13 +247,13 @@ findModule n = do
 
     path:rest -> do
       b <- io (doesFileExist path)
-      if b then return path else loop rest
+      if b then return (InFile path) else loop rest
 
     [] -> handleNotFound
 
   handleNotFound =
     case n of
-      m | m == preludeName -> io writePreludeContents
+      m | m == preludeName -> pure (InMem "Cryptol" preludeContents)
       _ -> moduleNotFound n =<< getSearchPath
 
   -- generate all possible search paths
@@ -350,7 +372,7 @@ getPrimMap  =
                   [ "Unable to find the prelude" ]
 
 -- | Load a module, be it a normal module or a functor instantiation.
-checkModule :: ImportSource -> FilePath -> P.Module PName -> ModuleM T.Module
+checkModule :: ImportSource -> ModulePath -> P.Module PName -> ModuleM T.Module
 checkModule isrc path m =
   case P.mInstance m of
     Nothing -> checkSingleModule T.tcModule isrc path m
@@ -364,7 +386,7 @@ checkModule isrc path m =
 checkSingleModule ::
   Act (P.Module Name) T.Module {- ^ how to check -} ->
   ImportSource                 {- ^ why are we loading this -} ->
-  FilePath                     {- path -} ->
+  ModulePath                   {- path -} ->
   P.Module PName               {- ^ module to check -} ->
   ModuleM T.Module
 checkSingleModule how isrc path m = do
@@ -374,8 +396,14 @@ checkSingleModule how isrc path m = do
   unless (notParamInstModName nm == thing (P.mName m))
          (moduleNameMismatch nm (mName m))
 
-  -- remove includes first
-  e   <- io (removeIncludesModule path m)
+  -- remove includes first; we only do this for actual files.
+  -- it is less clear what should happen for in-memory things, and since
+  -- this is a more-or-less obsolete feature, we are just not doing
+  -- ot for now.
+  e   <- case path of
+           InFile p -> io (removeIncludesModule p m)
+           InMem {} -> pure (Right m)
+
   nim <- case e of
            Right nim  -> return nim
            Left ierrs -> noIncludeErrors ierrs
@@ -493,6 +521,7 @@ genInferInput r prims params env = do
     , T.inpVars      = Map.map ifDeclSig (ifDecls env)
     , T.inpTSyns     = ifTySyns env
     , T.inpNewtypes  = ifNewtypes env
+    , T.inpAbstractTypes = ifAbstractTypes env
     , T.inpNameSeeds = seeds
     , T.inpMonoBinds = monoBinds
     , T.inpSolverConfig = cfg
