@@ -1,5 +1,5 @@
 -- |
--- Module      :  $Header$
+-- Module      :  Cryptol.Parser.Position
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
 -- License     :  BSD3
 -- Maintainer  :  cryptol@galois.com
@@ -13,8 +13,8 @@
 {-# LANGUAGE RecordWildCards #-}
 module Cryptol.Parser.Position where
 
-import           Data.Text.Lazy (Text)
-import qualified Data.Text.Lazy as T
+import           Data.Text(Text)
+import qualified Data.Text as T
 
 import GHC.Generics (Generic)
 import Control.DeepSeq
@@ -24,6 +24,7 @@ import Cryptol.Utils.PP
 data Located a  = Located { srcRange :: !Range, thing :: !a }
                   deriving (Eq, Show, Generic, NFData)
 
+
 data Position   = Position { line :: !Int, col :: !Int }
                   deriving (Eq, Ord, Show, Generic, NFData)
 
@@ -35,7 +36,7 @@ data Range      = Range { from   :: !Position
 -- | An empty range.
 --
 -- Caution: using this on the LHS of a use of rComb will cause the empty source
--- to propegate.
+-- to propagate.
 emptyRange :: Range
 emptyRange  = Range { from = start, to = start, source = "" }
 
@@ -65,14 +66,14 @@ instance Functor Located where
 --------------------------------------------------------------------------------
 
 instance PP Position where
-  ppPrec _ p = int (line p) <> colon <> int (col p)
+  ppPrec _ p = int (line p) <.> colon <.> int (col p)
 
 instance PP Range where
-  ppPrec _ r = text (source r) <> char ':'
-            <> pp (from r) <> text "--" <> pp (to r)
+  ppPrec _ r = text (source r) <.> char ':'
+            <.> pp (from r) <.> text "--" <.> pp (to r)
 
 instance PP a => PP (Located a) where
-  ppPrec _ l = parens (text "at" <+> pp (srcRange l) <> comma <+> pp (thing l))
+  ppPrec _ l = parens (text "at" <+> pp (srcRange l) <.> comma <+> pp (thing l))
 
 instance PPName a => PPName (Located a) where
   ppNameFixity  Located { .. } = ppNameFixity thing
@@ -118,4 +119,9 @@ instance AddLoc (Located a) where
 at :: (HasLoc l, AddLoc t) => l -> t -> t
 at l e = maybe e (addLoc e) (getLoc l)
 
+combLoc :: (a -> b -> c) -> Located a -> Located b -> Located c
+combLoc f l1 l2 = Located { srcRange = rComb (srcRange l1) (srcRange l2)
+                          , thing    = f (thing l1) (thing l2)
+                          }
+
 
