@@ -13,6 +13,7 @@
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE ViewPatterns #-}
+{-# LANGUAGE FlexibleInstances #-}
 
 module Cryptol.REPL.Monad (
     -- * REPL Monad
@@ -21,6 +22,7 @@ module Cryptol.REPL.Monad (
   , raise
   , stop
   , catch
+  , finally
   , rPutStrLn
   , rPutStr
   , rPrint
@@ -40,7 +42,8 @@ module Cryptol.REPL.Monad (
   , getTypeNames
   , getPropertyNames
   , getModNames
-  , LoadedModule(..), getLoadedMod, setLoadedMod, clearLoadedMod, setEditPath
+  , LoadedModule(..), getLoadedMod, setLoadedMod, clearLoadedMod
+  , setEditPath, getEditPath, clearEditPath
   , setSearchPath, prependSearchPath
   , getPrompt
   , shouldContinue
@@ -56,10 +59,11 @@ module Cryptol.REPL.Monad (
     -- ** Config Options
   , EnvVal(..)
   , OptionDescr(..)
-  , setUser, getUser, tryGetUser
+  , setUser, getUser, getKnownUser, tryGetUser
   , userOptions
   , getUserSatNum
   , getUserShowProverStats
+  , getUserProverValidate
 
     -- ** Configurable Output
   , getPutStr
@@ -93,7 +97,7 @@ import Cryptol.Utils.Logger(Logger, logPutStr, funLogger)
 import qualified Cryptol.Parser.AST as P
 import Cryptol.Symbolic (proverNames, lookupProver, SatNum(..))
 
-import Control.Monad (ap,unless,when,msum)
+import Control.Monad (ap,unless,when)
 import Control.Monad.Base
 import Control.Monad.IO.Class
 import Control.Monad.Trans.Control
@@ -101,7 +105,7 @@ import Data.Char (isSpace)
 import Data.IORef
     (IORef,newIORef,readIORef,modifyIORef,atomicModifyIORef)
 import Data.List (intercalate, isPrefixOf, unfoldr, sortBy)
-import Data.Maybe (catMaybes,mapMaybe)
+import Data.Maybe (catMaybes)
 import Data.Ord (comparing)
 import Data.Typeable (Typeable)
 import System.Directory (findExecutable)
@@ -119,15 +123,18 @@ import Prelude.Compat
 
 -- | This indicates what the user would like to work on.
 data LoadedModule = LoadedModule
-  { lName :: Maybe P.ModName -- ^ Working on this module.
-  , lPath :: FilePath        -- ^ Working on this file.
+  { lName :: Maybe P.ModName  -- ^ Working on this module.
+  , lPath :: M.ModulePath     -- ^ Working on this file.
   }
 
 -- | REPL RW Environment.
 data RW = RW
   { eLoadedMod   :: Maybe LoadedModule
     -- ^ This is the name of the currently "focused" module.
-    -- This is what we edit (:e) or reload (:r)
+    -- This is what we reload (:r)
+
+  , eEditFile :: Maybe FilePath
+    -- ^ This is what we edit (:e)
 
   , eContinue    :: Bool
     -- ^ Should we keep going when we encounter an error, or give up.
@@ -158,6 +165,7 @@ defaultRW isBatch l = do
   env <- M.initialModuleEnv
   return RW
     { eLoadedMod   = Nothing
+    , eEditFile    = Nothing
     , eContinue    = True
     , eIsBatch     = isBatch
     , eModuleEnv   = env
@@ -171,7 +179,37 @@ defaultRW isBatch l = do
 mkPrompt :: RW -> String
 mkPrompt rw
   | eIsBatch rw = ""
-  | otherwise   = maybe "cryptol" pretty (lName =<< eLoadedMod rw) ++ "> "
+  | detailedPrompt = withEdit ++ "> "
+  | otherwise      = modLn ++ "> "
+  where
+  detailedPrompt = False
+
+  modLn   =
+    case lName =<< eLoadedMod rw of
+      Nothing -> show (pp I.preludeName)
+      Just m
+        | M.isLoadedParamMod m (M.meLoadedModules (eModuleEnv rw)) ->
+                 modName ++ "(parameterized)"
+        | otherwise -> modName
+        where modName = pretty m
+
+  withFocus =
+    case eLoadedMod rw of
+      Nothing -> modLn
+      Just m ->
+        case (lName m, lPath m) of
+          (Nothing, M.InFile f) -> ":r to reload " ++ show f ++ "\n" ++ modLn
+          _ -> modLn
+
+  withEdit =
+    case eEditFile rw of
+      Nothing -> withFocus
+      Just e
+        | Just (M.InFile f) <- lPath <$> eLoadedMod rw
+        , f == e -> withFocus
+        | otherwise -> ":e to edit " ++ e ++ "\n" ++ withFocus
+
+
 
 -- REPL Monad ------------------------------------------------------------------
 
@@ -233,7 +271,7 @@ data REPLException
   | ModuleSystemError NameDisp M.ModuleError
   | EvalPolyError T.Schema
   | TypeNotTestable T.Type
-  | EvalInParamModule P.ModName [M.Name]
+  | EvalInParamModule [M.Name]
   | SBVError String
     deriving (Show,Typeable)
 
@@ -258,8 +296,9 @@ instance PP REPLException where
                          $$ text "Type:" <+> pp s
     TypeNotTestable t    -> text "The expression is not of a testable type."
                          $$ text "Type:" <+> pp t
-    EvalInParamModule _ xs ->
-      text "Expression depends on a module parameter:" <+> hsep (map pp xs)
+    EvalInParamModule xs ->
+      text "Expression depends on definitions from a parameterized module:"
+        $$ nest 2 (vcat (map pp xs))
     SBVError s           -> text "SBV error:" $$ text s
 
 -- | Raise an exception.
@@ -270,6 +309,9 @@ raise exn = io (X.throwIO exn)
 catch :: REPL a -> (REPLException -> REPL a) -> REPL a
 catch m k = REPL (\ ref -> rethrowEvalError (unREPL m ref) `X.catch` \ e -> unREPL (k e) ref)
 
+finally :: REPL a -> REPL b -> REPL a
+finally m1 m2 = REPL (\ref -> unREPL m1 ref `X.finally` unREPL m2 ref)
+
 
 rethrowEvalError :: IO a -> IO a
 rethrowEvalError m = run `X.catch` rethrow
@@ -303,25 +345,33 @@ modifyRW_ f = REPL (\ ref -> modifyIORef ref f)
 getPrompt :: REPL String
 getPrompt  = mkPrompt `fmap` getRW
 
-
 clearLoadedMod :: REPL ()
 clearLoadedMod = do modifyRW_ (\rw -> rw { eLoadedMod = upd <$> eLoadedMod rw })
                     updateREPLTitle
   where upd x = x { lName = Nothing }
 
--- | Set the name of the currently focused file, edited by @:e@ and loaded via
--- @:r@.
+-- | Set the name of the currently focused file, loaded via @:r@.
 setLoadedMod :: LoadedModule -> REPL ()
 setLoadedMod n = do
   modifyRW_ (\ rw -> rw { eLoadedMod = Just n })
   updateREPLTitle
 
-setEditPath :: FilePath -> REPL ()
-setEditPath p = setLoadedMod LoadedModule { lName = Nothing, lPath = p }
-
 getLoadedMod :: REPL (Maybe LoadedModule)
 getLoadedMod  = eLoadedMod `fmap` getRW
 
+
+
+-- | Set the path for the ':e' command.
+-- Note that this does not change the focused module (i.e., what ":r" reloads)
+setEditPath :: FilePath -> REPL ()
+setEditPath p = modifyRW_ $ \rw -> rw { eEditFile = Just p }
+
+getEditPath :: REPL (Maybe FilePath)
+getEditPath = eEditFile <$> getRW
+
+clearEditPath :: REPL ()
+clearEditPath = modifyRW_ $ \rw -> rw { eEditFile = Nothing }
+
 setSearchPath :: [FilePath] -> REPL ()
 setSearchPath path = do
   me <- getModuleEnv
@@ -368,35 +418,19 @@ getLetEnabled = fmap eLetEnabled getRW
 validEvalContext :: T.FreeVars a => a -> REPL ()
 validEvalContext a =
   do me <- eModuleEnv <$> getRW
-     case M.meFocusedModule me of
-       Nothing -> return ()
-       Just fm ->
-         case M.lookupModule fm me of
-           Just lm
-             | Set.null tps && Set.null vps -> return ()
-             | Just xs <- check (T.freeVars a) ->
-                              raise $ EvalInParamModule fm xs
-             | otherwise -> return ()
-             where
-             m = M.lmModule lm
+     let ds      = T.freeVars a
+         badVals = foldr badName Set.empty (T.valDeps ds)
+         bad     = foldr badName badVals (T.tyDeps ds)
 
-             tps = Set.fromList $ map T.mtpParam $ Map.elems $ T.mParamTypes m
-             vps = Map.keysSet (T.mParamFuns m)
+         badName nm bs =
+           case M.nameInfo nm of
+             M.Declared m _
+               | M.isLoadedParamMod m (M.meLoadedModules me) -> Set.insert nm bs
+             _ -> bs
 
-             bad ds | not $ null badTP = Just $ mapMaybe T.tpName badTP
-                    | not $ null badVP = Just badVP
-                    | otherwise = Nothing
-               where
-               badTP = Set.toList $ Set.intersection tps (T.tyParams ds)
-               badVP = Set.toList $ Set.intersection vps (T.valDeps ds)
+     unless (Set.null bad) $
+       raise (EvalInParamModule (Set.toList bad))
 
-             check ds   = msum (bad ds : map badSub (Set.toList (T.valDeps ds)))
-             badSub x   = maybe Nothing bad (Map.lookup x deps)
-             deps       = T.moduleDeps m
-
-           Nothing ->
-             panic "getEvalEnabled" ["The focused module is not loaded."
-                                    , show fm ]
 
 
 -- | Update the title
@@ -650,10 +684,41 @@ getUser name = do
     Just ev -> return ev
     Nothing -> panic "[REPL] getUser" ["option `" ++ name ++ "` does not exist"]
 
+getKnownUser :: IsEnvVal a => String -> REPL a
+getKnownUser x = fromEnvVal <$> getUser x
+
+class IsEnvVal a where
+  fromEnvVal :: EnvVal -> a
+
+instance IsEnvVal Bool where
+  fromEnvVal x = case x of
+                   EnvBool b -> b
+                   _         -> badIsEnv "Bool"
+
+instance IsEnvVal Int where
+  fromEnvVal x = case x of
+                   EnvNum b -> b
+                   _         -> badIsEnv "Num"
+
+instance IsEnvVal (String,[String]) where
+  fromEnvVal x = case x of
+                   EnvProg b bs -> (b,bs)
+                   _            -> badIsEnv "Prog"
+
+instance IsEnvVal String where
+  fromEnvVal x = case x of
+                   EnvString b -> b
+                   _           -> badIsEnv "String"
+
+badIsEnv :: String -> a
+badIsEnv x = panic "fromEnvVal" [ "[REPL] Expected a `" ++ x ++ "` value." ]
+
+
 getUserShowProverStats :: REPL Bool
-getUserShowProverStats =
-  do EnvBool yes <- getUser "prover-stats"
-     return yes
+getUserShowProverStats = getKnownUser "prover-stats"
+
+getUserProverValidate :: REPL Bool
+getUserProverValidate = getKnownUser "prover-validate"
 
 -- Environment Options ---------------------------------------------------------
 
@@ -728,7 +793,7 @@ userOptions  = mkOptionMap
     "Enable type-checker debugging output." $
     \case EnvNum n -> do me <- getModuleEnv
                          let cfg = M.meSolverConfig me
-                         setModuleEnv me { M.meSolverConfig = cfg{ T.solverVerbose = fromIntegral n } }
+                         setModuleEnv me { M.meSolverConfig = cfg{ T.solverVerbose = n } }
           _        -> return ()
   , OptionDescr "core-lint" (EnvBool False)
     noCheck
@@ -741,6 +806,12 @@ userOptions  = mkOptionMap
 
   , simpleOpt "prover-stats" (EnvBool True) noCheck
     "Enable prover timing statistics."
+
+  , simpleOpt "prover-validate" (EnvBool False) noCheck
+    "Validate :sat examples and :prove counter-examples for correctness."
+
+  , simpleOpt "show-examples" (EnvBool True) noCheck
+    "Print the (counter) example after :sat or :prove"
   ]
 
 
@@ -789,7 +860,7 @@ checkSatNum val = case val of
 
 getUserSatNum :: REPL SatNum
 getUserSatNum = do
-  EnvString s <- getUser "satNum"
+  s <- getKnownUser "satNum"
   case s of
     "all"                     -> return AllSat
     _ | Just n <- readMaybe s -> return (SomeSat n)
@@ -800,7 +871,7 @@ getUserSatNum = do
 
 whenDebug :: REPL () -> REPL ()
 whenDebug m = do
-  EnvBool b <- getUser "debug"
+  b <- getKnownUser "debug"
   when b m
 
 -- Smoke Testing ---------------------------------------------------------------
@@ -830,3 +901,4 @@ z3exists = do
   case mPath of
     Nothing -> return (Just Z3NotFound)
     Just _  -> return Nothing
+
