@@ -9,17 +9,19 @@
 -- This module generates random values for Cryptol types.
 
 {-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 module Cryptol.Testing.Random where
 
-import Cryptol.Eval.Monad     (ready,EvalOpts)
+import Cryptol.Eval.Monad     (ready,runEval,EvalOpts)
 import Cryptol.Eval.Value     (BV(..),Value,GenValue(..),SeqMap(..), WordValue(..), BitWord(..))
 import qualified Cryptol.Testing.Concrete as Conc
-import Cryptol.TypeCheck.AST  (Type(..),TCon(..),TC(..),tNoUser)
+import Cryptol.TypeCheck.AST  (Type(..), TCon(..), TC(..), tNoUser, tIsFun)
 import Cryptol.TypeCheck.SimpType(tRebuild')
 
 import Cryptol.Utils.Ident    (Ident)
+import Cryptol.Utils.Panic    (panic)
 
-import Control.Monad          (forM)
+import Control.Monad          (forM,join)
 import Data.List              (unfoldr, genericTake, genericIndex)
 import System.Random          (RandomGen, split, random, randomR)
 import qualified Data.Sequence as Seq
@@ -47,6 +49,57 @@ runOneTest evOpts fun argGens sz g0 = do
   result <- Conc.runOneTest evOpts fun args
   return (result, g1)
 
+returnOneTest :: RandomGen g
+           => EvalOpts -- ^ How to evaluate things
+           -> Value    -- ^ Function to be used to calculate tests
+           -> [Gen g Bool BV Integer] -- ^ Argument generators
+           -> Integer -- ^ Size
+           -> g -- ^ Initial random state
+           -> IO ([Value], Value, g) -- ^ Arguments, result, and new random state
+returnOneTest evOpts fun argGens sz g0 =
+  do let (args, g1) = foldr mkArg ([], g0) argGens
+         mkArg argGen (as, g) = let (a, g') = argGen sz g in (a:as, g')
+     result <- runEval evOpts (go fun args)
+     return (args, result, g1)
+   where
+     go (VFun f) (v : vs) = join (go <$> (f (ready v)) <*> pure vs)
+     go (VFun _) [] = panic "Cryptol.Testing.Random" ["Not enough arguments to function while generating tests"]
+     go _ (_ : _) = panic "Cryptol.Testing.Random" ["Too many arguments to function while generating tests"]
+     go v [] = return v
+
+
+-- | Return a collection of random tests.
+returnTests :: RandomGen g
+         => g -- ^ The random generator state
+         -> EvalOpts -- ^ How to evaluate things
+         -> [Gen g Bool BV Integer] -- ^ Generators for the function arguments
+         -> Value -- ^ The function itself
+         -> Int -- ^ How many tests?
+         -> IO [([Value], Value)] -- ^ A list of pairs of random arguments and computed outputs
+returnTests g evo gens fun num = go gens g 0
+  where
+    go args g0 n
+      | n >= num = return []
+      | otherwise =
+        do let sz = toInteger (div (100 * (1 + n)) num)
+           (inputs, output, g1) <- returnOneTest evo fun args sz g0
+           more <- go args g1 (n + 1)
+           return ((inputs, output) : more)
+
+{- | Given a (function) type, compute generators for the function's
+arguments. This is like @testableType@, but allows the result to be
+any finite type instead of just @Bit@. -}
+dumpableType :: forall g. RandomGen g => Type -> Maybe [Gen g Bool BV Integer]
+dumpableType ty =
+  case tIsFun ty of
+    Just (t1, t2) ->
+      do g  <- randomValue t1
+         as <- testableType t2
+         return (g : as)
+    Nothing ->
+      do (_ :: Gen g Bool BV Integer) <- randomValue ty
+         return []
+
 {- | Given a (function) type, compute generators for
 the function's arguments. Currently we do not support polymorphic functions.
 In principle, we could apply these to random types, and test the results. -}
