@@ -33,7 +33,6 @@ import Cryptol.Parser.Position(Located)
 import Cryptol.ModuleSystem.Name
 import Cryptol.ModuleSystem.Exports(ExportSpec(..)
                                    , isExportedBind, isExportedType)
-import Cryptol.Prims.Syntax
 import Cryptol.Parser.AST ( Selector(..),Pragma(..)
                           , Import(..), ImportSpec(..), ExportType(..)
                           , Fixity(..))
@@ -59,6 +58,7 @@ data Module = Module { mName        :: !ModName
                        -- of a module.
 
                      , mNewtypes         :: Map Name Newtype
+                     , mPrimTypes        :: Map Name AbstractType
                      , mParamTypes       :: Map Name ModTParam
                      , mParamConstraints :: [Located Prop]
                      , mParamFuns        :: Map Name ModVParam
@@ -105,6 +105,7 @@ data Expr   = EList [Expr] Type         -- ^ List value (with type of elements)
             | ETuple [Expr]             -- ^ Tuple value
             | ERec [(Ident,Expr)]       -- ^ Record value
             | ESel Expr Selector        -- ^ Elimination for tuple/record/list
+            | ESet Expr Selector Expr   -- ^ Change the value of a field.
 
             | EIf Expr Expr Expr        -- ^ If-then-else
             | EComp Type Type Expr [[Match]]
@@ -210,6 +211,8 @@ instance PP (WithNames Expr) where
 
       ESel e sel    -> ppWP 4 e <+> text "." <.> pp sel
 
+      ESet e sel v  -> braces (pp e <+> "|" <+> pp sel <+> "=" <+> pp v)
+
       EIf e1 e2 e3  -> optParens (prec > 0)
                     $ sep [ text "if"   <+> ppW e1
                           , text "then" <+> ppW e2
