@@ -6,16 +6,23 @@
 -- Stability   :  provisional
 -- Portability :  portable
 {-# Language OverloadedStrings #-}
-module Cryptol.TypeCheck.Instantiate (instantiateWith) where
+module Cryptol.TypeCheck.Instantiate
+  ( instantiateWith
+  , TypeArg(..)
+  , uncheckedTypeArg
+  , MaybeCheckedType(..)
+  ) where
 
 import Cryptol.ModuleSystem.Name (nameIdent)
 import Cryptol.TypeCheck.AST
 import Cryptol.TypeCheck.Monad
-import Cryptol.TypeCheck.Subst (listSubst,apSubst)
+import Cryptol.TypeCheck.Subst (listParamSubst, apSubst)
+import Cryptol.TypeCheck.Kind(checkType)
 import Cryptol.TypeCheck.Error
 import Cryptol.Parser.Position (Located(..))
 import Cryptol.Utils.Ident (Ident)
 import Cryptol.Utils.Panic(panic)
+import qualified Cryptol.Parser.AST as P
 
 import Control.Monad(zipWithM)
 import Data.Function (on)
@@ -25,9 +32,37 @@ import Data.Either(partitionEithers)
 import qualified Data.Set as Set
 
 
+data TypeArg = TypeArg
+  { tyArgName :: Maybe (Located Ident)
+  , tyArgType :: MaybeCheckedType
+  }
 
-instantiateWith :: Name -> Expr -> Schema -> [Located (Maybe Ident,Type)]
-                -> InferM (Expr,Type)
+uncheckedTypeArg :: P.TypeInst Name -> TypeArg
+uncheckedTypeArg a =
+  case a of
+    P.NamedInst x ->
+      TypeArg { tyArgName = Just (P.name x), tyArgType = Unchecked (P.value x) }
+    P.PosInst t ->
+      TypeArg { tyArgName = Nothing, tyArgType = Unchecked t }
+
+
+data MaybeCheckedType = Checked Type | Unchecked (P.Type Name)
+
+
+
+checkTyParam :: TVarSource -> Kind -> MaybeCheckedType -> InferM Type
+checkTyParam src k mb =
+  case mb of
+    Checked t
+      | k == k'   -> pure t
+      | otherwise -> do recordError (KindMismatch k k')
+                        newType src k
+        where k' = kindOf t
+    Unchecked t -> checkType t (Just k)
+
+
+
+instantiateWith :: Name -> Expr -> Schema -> [TypeArg] -> InferM (Expr,Type)
 instantiateWith nm e s ts
   | null named      = instantiateWithPos nm e s positional
   | null positional = instantiateWithNames nm e s named
@@ -36,27 +71,25 @@ instantiateWith nm e s ts
 
   where
   (named,positional) = partitionEithers (map classify ts)
-  classify t = case thing t of
-                 (Just n,ty)  -> Left t { thing = (n,ty) }
-                 (Nothing,ty) -> Right ty
+  classify t = case tyArgName t of
+                 Just n  -> Left n { thing = (thing n, tyArgType t) }
+                 Nothing -> Right (tyArgType t)
 
 
-instantiateWithPos :: Name -> Expr -> Schema -> [Type] -> InferM (Expr,Type)
+
+instantiateWithPos ::
+  Name -> Expr -> Schema -> [MaybeCheckedType] -> InferM (Expr,Type)
 instantiateWithPos nm e (Forall as ps t) ts =
   do su <- makeSu (1::Int) [] as ts
      doInst su e ps t
   where
   isNamed q = isJust (tpName q)
 
-  makeSu n su (q : qs) (ty : tys)
+  makeSu n su (q : qs) (mbty : tys)
     | not (isNamed q) = do r <- unnamed n q
-                           makeSu (n+1) (r : su) qs (ty : tys)
-    | k1 == k2        = makeSu (n+1) ((q, ty) : su) qs tys
-    | otherwise       = do recordError (KindMismatch k1 k2)
-                           r <- unnamed n q
-                           makeSu (n+1) (r : su) qs tys
-      where k1 = kindOf q
-            k2 = kindOf ty
+                           makeSu (n+1) (r : su) qs (mbty : tys)
+    | otherwise = do ty <- checkTyParam (TypeParamInstPos nm n) (kindOf q) mbty
+                     makeSu (n+1) ((q,ty) : su) qs tys
 
   makeSu _ su [] []       = return (reverse su)
   makeSu n su (q : qs) [] = do r <- unnamed n q
@@ -79,7 +112,6 @@ instantiateWithPos nm e (Forall as ps t) ts =
 
 
 
-
 {- | Instantiate an expression of the given polymorphic type.
 The arguments that are provided will be instantiated as requested,
 the rest will be instantiated with fresh type variables.
@@ -90,7 +122,8 @@ EProofApp (ETApp e t)
   - There will be one `ETApp t` for each insantiated type parameter;
   - there will be one `EProofApp` for each constraint on the schema;
 -}
-instantiateWithNames :: Name -> Expr -> Schema -> [Located (Ident,Type)]
+instantiateWithNames ::
+  Name -> Expr -> Schema -> [Located (Ident,MaybeCheckedType)]
                      -> InferM (Expr,Type)
 instantiateWithNames nm e (Forall as ps t) xs =
   do sequence_ repeatedParams
@@ -109,13 +142,7 @@ instantiateWithNames nm e (Forall as ps t) xs =
                    Just na -> TypeParamInstNamed nm (nameIdent na)
                    Nothing -> TypeParamInstPos nm n
        ty <- case lkp =<< tpName x of
-               Just lty
-                 | k1 == k   -> return ty
-                 | otherwise -> inRange (srcRange lty) $
-                                  do recordError (KindMismatch k k1)
-                                     newType src k
-                  where ty = snd (thing lty)
-                        k1 = kindOf ty
+               Just lty -> checkTyParam src k (snd (thing lty))
                Nothing -> newType src k
        return (x, ty)
 
@@ -147,7 +174,7 @@ instantiateWithNames nm e (Forall as ps t) xs =
 
 doInst :: [(TParam, Type)] -> Expr -> [Prop] -> Type -> InferM (Expr,Type)
 doInst su' e ps t =
-  do let su = listSubst [ (tpVar tp, ty) | (tp, ty) <- su' ]
+  do let su = listParamSubst su'
      newGoals (CtInst e) (map (apSubst su) ps)
      let t1 = apSubst su t
 
