@@ -3,7 +3,7 @@ module Cryptol.TypeCheck.SimpType where
 
 import Control.Applicative((<|>))
 import Cryptol.TypeCheck.Type hiding
-  (tSub,tMul,tDiv,tMod,tExp,tMin,tCeilDiv,tCeilMod,tLenFromThen,tLenFromThenTo)
+  (tSub,tMul,tDiv,tMod,tExp,tMin,tCeilDiv,tCeilMod,tLenFromThenTo)
 import Cryptol.TypeCheck.TypePat
 import Cryptol.TypeCheck.Solver.InfNat
 import Control.Monad(msum,guard)
@@ -18,31 +18,33 @@ tRebuild' withUser = go
       TUser x xs t
         | withUser  -> TUser x xs (go t)
         | otherwise -> go t
-      TVar _       -> ty
-      TRec xs      -> TRec [ (x,go y) | (x,y) <- xs ]
-      TCon tc ts ->
-        case (tc, map go ts) of
-          (TF f, ts') ->
-            case (f,ts') of
-              (TCAdd,[x,y]) -> tAdd x y
-              (TCSub,[x,y]) -> tSub x y
-              (TCMul,[x,y]) -> tMul x y
-              (TCExp,[x,y]) -> tExp x y
-              (TCDiv,[x,y]) -> tDiv x y
-              (TCMod,[x,y]) -> tMod x y
-              (TCMin,[x,y]) -> tMin x y
-              (TCMax,[x,y]) -> tMax x y
-              (TCWidth,[x]) -> tWidth x
-              (TCCeilDiv,[x,y]) -> tCeilDiv x y
-              (TCCeilMod,[x,y]) -> tCeilMod x y
-              (TCLenFromThen,[x,y,z]) -> tLenFromThen x y z
-              (TCLenFromThenTo,[x,y,z]) -> tLenFromThenTo x y z
-              _ -> TCon tc ts
-          (_,ts') -> TCon tc ts'
+      TVar _        -> ty
+      TRec xs       -> TRec [ (x, go y) | (x, y) <- xs ]
+      TCon tc ts    -> tCon tc (map go ts)
 
 tRebuild :: Type -> Type
 tRebuild = tRebuild' True
 
+tCon :: TCon -> [Type] -> Type
+tCon tc ts =
+  case tc of
+    TF f ->
+      case (f, ts) of
+        (TCAdd, [x, y]) -> tAdd x y
+        (TCSub, [x, y]) -> tSub x y
+        (TCMul, [x, y]) -> tMul x y
+        (TCExp, [x, y]) -> tExp x y
+        (TCDiv, [x, y]) -> tDiv x y
+        (TCMod, [x, y]) -> tMod x y
+        (TCMin, [x, y]) -> tMin x y
+        (TCMax, [x, y]) -> tMax x y
+        (TCWidth, [x]) -> tWidth x
+        (TCCeilDiv, [x, y]) -> tCeilDiv x y
+        (TCCeilMod, [x, y]) -> tCeilMod x y
+        (TCLenFromThenTo, [x, y, z]) -> tLenFromThenTo x y z
+        _ -> TCon tc ts
+    _ -> TCon tc ts
+
 -- Normal: constants to the left
 tAdd :: Type -> Type -> Type
 tAdd x y
@@ -280,12 +282,6 @@ tWidth x
   | Just t <- tOp TCWidth (total (op1 nWidth)) [x] = t
   | otherwise = tf1 TCWidth x
 
-tLenFromThen :: Type -> Type -> Type -> Type
-tLenFromThen x y z
-  | Just t <- tOp TCLenFromThen (op3 nLenFromThen) [x,y,z] = t
-  -- XXX: rules?
-  | otherwise = tf3 TCLenFromThen x y z
-
 tLenFromThenTo :: Type -> Type -> Type -> Type
 tLenFromThenTo x y z
   | Just t <- tOp TCLenFromThenTo (op3 nLenFromThenTo) [x,y,z] = t
