@@ -18,6 +18,7 @@ module Cryptol.TypeCheck.Subst
   , (@@)
   , defaultingSubst
   , listSubst
+  , listParamSubst
   , isEmptySubst
   , FVS(..)
   , apSubstMaybe
@@ -43,69 +44,115 @@ import qualified Cryptol.TypeCheck.SimpleSolver as Simp
 import Cryptol.Utils.Panic(panic)
 import Cryptol.Utils.Misc(anyJust)
 
-{- | Type 'Subst' has an invariant on its 'suMap' component: If there
-is a mapping from @TVFree _ _ tps _@ to a type @t@, then @t@ must not
-mention (directly or indirectly) any type parameter that is not in
-@tps@. In particular, if @t@ contains a variable @TVFree _ _ tps2 _@,
-then @tps2@ must be a subset of @tps@. This ensures that applying the
-substitution will not permit any type parameter to escape from its
-scope. -}
+-- | A 'Subst' value represents a substitution that maps each 'TVar'
+-- to a 'Type'.
+--
+-- Invariant: If there is a mapping from @TVFree _ _ tps _@ to a type
+-- @t@, then @t@ must not mention (directly or indirectly) any type
+-- parameter that is not in @tps@. In particular, if @t@ contains a
+-- variable @TVFree _ _ tps2 _@, then @tps2@ must be a subset of
+-- @tps@. This ensures that applying the substitution will not permit
+-- any type parameter to escape from its scope.
 
-data Subst = S { suMap :: !(Map.Map TVar Type)
+data Subst = S { suFreeMap :: !(IntMap.IntMap (TVar, Type))
+               , suBoundMap :: !(IntMap.IntMap (TVar, Type))
                , suDefaulting :: !Bool
                }
                   deriving Show
 
 emptySubst :: Subst
-emptySubst = S { suMap = Map.empty, suDefaulting = False }
+emptySubst =
+  S { suFreeMap = IntMap.empty
+    , suBoundMap = IntMap.empty
+    , suDefaulting = False
+    }
 
 singleSubst :: TVar -> Type -> Subst
-singleSubst x t = S { suMap = Map.singleton x t, suDefaulting = False }
+singleSubst v@(TVFree i _ _tps _) t =
+  S { suFreeMap = IntMap.singleton i (v, t)
+    , suBoundMap = IntMap.empty
+    , suDefaulting = False
+    }
+singleSubst v@(TVBound tp) t =
+  S { suFreeMap = IntMap.empty
+    , suBoundMap = IntMap.singleton (tpUnique tp) (v, t)
+    , suDefaulting = False
+    }
 
 (@@) :: Subst -> Subst -> Subst
 s2 @@ s1
-  | Map.null (suMap s2) =
+  | isEmptySubst s2 =
     if suDefaulting s1 || not (suDefaulting s2) then
       s1
     else
       s1{ suDefaulting = True }
 
-s2 @@ s1 = S { suMap = Map.map (apSubst s2) (suMap s1) `Map.union` suMap s2
-             , suDefaulting = suDefaulting s1 || suDefaulting s2
-             }
-
+s2 @@ s1 =
+  S { suFreeMap = IntMap.map (fmap (apSubst s2)) (suFreeMap s1) `IntMap.union` suFreeMap s2
+    , suBoundMap = IntMap.map (fmap (apSubst s2)) (suBoundMap s1) `IntMap.union` suBoundMap s2
+    , suDefaulting = suDefaulting s1 || suDefaulting s2
+    }
 
+-- | A defaulting substitution maps all otherwise-unmapped free
+-- variables to a kind-appropriate default type (@Bit@ for value types
+-- and @0@ for numeric types).
 defaultingSubst :: Subst -> Subst
 defaultingSubst s = s { suDefaulting = True }
 
 -- | Makes a substitution out of a list.
 -- WARNING: We do not validate the list in any way, so the caller should
 -- ensure that we end up with a valid (e.g., idempotent) substitution.
-listSubst :: [(TVar,Type)] -> Subst
+listSubst :: [(TVar, Type)] -> Subst
 listSubst xs
   | null xs   = emptySubst
-  | otherwise = S { suMap = Map.fromList xs, suDefaulting = False }
+  | otherwise = S { suFreeMap = IntMap.fromList frees
+                  , suBoundMap = IntMap.fromList bounds
+                  , suDefaulting = False }
+  where
+    (frees, bounds) = partitionEithers (map classify xs)
+    classify x =
+      case fst x of
+        TVFree i _ _ _ -> Left (i, x)
+        TVBound tp -> Right (tpUnique tp, x)
+
+-- | Makes a substitution out of a list.
+-- WARNING: We do not validate the list in any way, so the caller should
+-- ensure that we end up with a valid (e.g., idempotent) substitution.
+listParamSubst :: [(TParam, Type)] -> Subst
+listParamSubst xs
+  | null xs   = emptySubst
+  | otherwise = S { suFreeMap = IntMap.empty
+                  , suBoundMap = IntMap.fromList bounds
+                  , suDefaulting = False }
+  where
+    bounds = [ (tpUnique tp, (TVBound tp, t)) | (tp, t) <- xs ]
 
 isEmptySubst :: Subst -> Bool
-isEmptySubst su = Map.null $ suMap su
+isEmptySubst su = IntMap.null (suFreeMap su) && IntMap.null (suBoundMap su)
 
 -- Returns the empty set if this is a defaulting substitution
 substBinds :: Subst -> Set TVar
 substBinds su
   | suDefaulting su = Set.empty
-  | otherwise       = Map.keysSet $ suMap su
+  | otherwise       = Set.fromList (map fst (assocsSubst su))
 
-substToList :: Subst -> [(TVar,Type)]
+substToList :: Subst -> [(TVar, Type)]
 substToList s
   | suDefaulting s = panic "substToList" ["Defaulting substitution."]
-  | otherwise = Map.toList (suMap s)
+  | otherwise = assocsSubst s
+
+assocsSubst :: Subst -> [(TVar, Type)]
+assocsSubst s = frees ++ bounds
+  where
+    frees = IntMap.elems (suFreeMap s)
+    bounds = IntMap.elems (suBoundMap s)
 
 instance PP (WithNames Subst) where
   ppPrec _ (WithNames s mp)
     | null els  = text "(empty substitution)"
     | otherwise = text "Substitution:" $$ nest 2 (vcat (map pp1 els))
     where pp1 (x,t) = ppWithNames mp x <+> text "=" <+> ppWithNames mp t
-          els       = Map.toList (suMap s)
+          els       = assocsSubst s
 
 instance PP Subst where
   ppPrec n = ppWithNamesPrec IntMap.empty n
@@ -121,28 +168,9 @@ apSubstMaybe su ty =
     TCon t ts ->
       do ss <- anyJust (apSubstMaybe su) ts
          case t of
-
-           TF f ->
-             Just $!
-             case (f,ss) of
-               (TCAdd,[t1,t2])               -> Simp.tAdd t1 t2
-               (TCSub,[t1,t2])               -> Simp.tSub t1 t2
-               (TCMul,[t1,t2])               -> Simp.tMul t1 t2
-               (TCDiv,[t1,t2])               -> Simp.tDiv t1 t2
-               (TCMod,[t1,t2])               -> Simp.tMod t1 t2
-               (TCExp,[t1,t2])               -> Simp.tExp t1 t2
-               (TCMin,[t1,t2])               -> Simp.tMin t1 t2
-               (TCMax,[t1,t2])               -> Simp.tMax t1 t2
-               (TCWidth,[t1])                -> Simp.tWidth t1
-               (TCCeilDiv,[t1,t2])           -> Simp.tCeilDiv t1 t2
-               (TCCeilMod,[t1,t2])           -> Simp.tCeilMod t1 t2
-               (TCLenFromThen,[t1,t2,t3])    -> Simp.tLenFromThen t1 t2 t3
-               (TCLenFromThenTo,[t1,t2,t3])  -> Simp.tLenFromThenTo t1 t2 t3
-               _ -> panic "apSubstMaybe" ["Unexpected type function", show t]
-
-           PC _ ->Just $! Simp.simplify Map.empty (TCon t ss)
-
-           _ -> return (TCon t ss)
+           TF _ -> Just $! Simp.tCon t ss
+           PC _ -> Just $! Simp.simplify Map.empty (TCon t ss)
+           _    -> Just (TCon t ss)
 
     TUser f ts t  -> do t1 <- apSubstMaybe su t
                         return (TUser f (map (apSubst su) ts) t1)
@@ -151,10 +179,18 @@ apSubstMaybe su ty =
                            return (x,t1)
     TVar x -> applySubstToVar su x
 
+lookupSubst :: TVar -> Subst -> Maybe Type
+lookupSubst x su =
+  fmap snd $
+  case x of
+    TVFree i _ _ _ -> IntMap.lookup i (suFreeMap su)
+    TVBound tp -> IntMap.lookup (tpUnique tp) (suBoundMap su)
 
 applySubstToVar :: Subst -> TVar -> Maybe Type
 applySubstToVar su x =
-  case Map.lookup x (suMap su) of
+  case lookupSubst x su of
+    -- For a defaulting substitution, we must recurse in order to
+    -- replace unmapped free vars with default types.
     Just t  -> Just (if suDefaulting su then apSubst su t else t)
     Nothing
       | suDefaulting su -> Just $! defaultFreeVar x
@@ -264,6 +300,7 @@ instance TVars Expr where
 
         ETuple es     -> ETuple (map go es)
         ERec fs       -> ERec [ (f, go e) | (f,e) <- fs ]
+        ESet e x v    -> ESet (go e) x (go v)
         EList es t    -> EList (map go es) (apSubst su t)
         ESel e s      -> ESel (go e) s
         EComp len t e mss -> EComp (apSubst su len) (apSubst su t) (go e) (apSubst su mss)
